'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.generatePushID = undefined;
exports.encode = encode;
exports.decode = decode;
exports.newGlobalID = newGlobalID;
exports.typeFromGlobalID = typeFromGlobalID;
exports.isGlobalID = isGlobalID;

var _invariant = require('./invariant');

var _invariant2 = _interopRequireDefault(_invariant);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/* Source:
 * https://www.firebase.com/blog/2015-02-11-firebase-unique-identifiers.html
 * https://gist.github.com/mikelehen/3596a30bd69384624c11
 */

/**
 * Fancy ID generator that creates 20-character string identifiers with the
 * following properties:
 *
 * 1. They're based on timestamp so that they sort *after* any existing ids.
 * 2. They contain 72-bits of random data after the timestamp so that IDs
 *    won't collide with other clients' IDs.
 * 3. They sort *lexicographically* (so the timestamp is converted to characters
 *    that will sort properly).
 * 4. They're monotonically increasing.  Even if you generate more than one in
 *    the same timestamp, the latter ones will sort after the former ones.
 *    We do this by using the previous random bits but "incrementing" them by 1
 *    (only in the case of a timestamp collision).
 */

var generatePushID = exports.generatePushID = function () {
  // Modeled after base64 web-safe chars, but ordered by ASCII.
  var PUSH_CHARS = '-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz';

  // Timestamp of last push, used to prevent local collisions if you push twice
  // in one ms.
  var lastPushTime = 0;

  // We generate 72-bits of randomness which get turned into 12 characters and
  // appended to the timestamp to prevent collisions with other clients.
  // We store the last characters we generated because in the event of a
  // collision, we'll use those same characters except "incremented" by one.
  var lastRandChars = [];

  return function () {
    var now = new Date().getTime();
    var duplicateTime = now === lastPushTime;
    lastPushTime = now;

    var timeStampChars = new Array(8);
    for (var i = 7; i >= 0; i--) {
      timeStampChars[i] = PUSH_CHARS.charAt(now % 64);
      // NOTE: Can't use << here because javascript will convert to int and
      // lose the upper bits.
      now = Math.floor(now / 64);
    }
    (0, _invariant2.default)(now === 0, 'We should have converted the entire timestamp.');

    var id = timeStampChars.join('');

    if (!duplicateTime) {
      for (var _i = 0; _i < 12; _i++) {
        lastRandChars[_i] = Math.floor(Math.random() * 64);
      }
    } else {
      // If the timestamp hasn't changed since last push, use the same random
      // number, except incremented by 1.
      var _i2 = void 0;
      for (_i2 = 11; _i2 >= 0 && lastRandChars[_i2] === 63; _i2--) {
        lastRandChars[_i2] = 0;
      }
      lastRandChars[_i2]++;
    }
    for (var _i3 = 0; _i3 < 12; _i3++) {
      id += PUSH_CHARS.charAt(lastRandChars[_i3]);
    }
    (0, _invariant2.default)(id.length === 20, 'Length should be 20.');

    return id;
  };
}();


var ENCODING_MAP = { '0': 'q', '1': 'r', '2': 's', '3': 't', '4': 'u', '5': 'v', '6': 'w',
  '7': 'x', '8': 'y', '9': 'z', _: '0',
  a: '1', b: '2', c: '3', d: '4', e: '5', f: '6', g: '7',
  h: '8', i: '9', j: 'A', k: 'B', l: 'C', m: 'D', n: 'E', o: 'F', p: 'G',
  q: 'H', r: 'I', s: 'J', t: 'K', u: 'L', v: 'M', w: 'N', x: 'O', y: 'P',
  z: 'Q', A: 'R', B: 'S', C: 'T', D: 'U', E: 'V', F: 'W', G: 'X', H: 'Y',
  I: 'Z', J: '_', K: 'a', L: 'b', M: 'c', N: 'd', O: 'e', P: 'f', Q: 'g',
  R: 'h', S: 'i', T: 'j', U: 'k', V: 'l', W: 'm', X: 'n', Y: 'o', Z: 'p' };

var DECODING_MAP = { '0': '_', '1': 'a', '2': 'b', '3': 'c', '4': 'd', '5': 'e', '6': 'f',
  '7': 'g', '8': 'h', '9': 'i',
  q: '0', r: '1', s: '2', t: '3', u: '4', v: '5', w: '6', x: '7',
  y: '8', z: '9', A: 'j', B: 'k', C: 'l', D: 'm', E: 'n', F: 'o', G: 'p',
  H: 'q', I: 'r', J: 's', K: 't', L: 'u', M: 'v', N: 'w', O: 'x', P: 'y',
  Q: 'z', R: 'A', S: 'B', T: 'C', U: 'D', V: 'E', W: 'F', X: 'G', Y: 'H',
  Z: 'I', _: 'J', a: 'K', b: 'L', c: 'M', d: 'N', e: 'O', f: 'P', g: 'Q',
  h: 'R', i: 'S', j: 'T', k: 'U', l: 'V', m: 'W', n: 'X', o: 'Y', p: 'Z' };

function encode(type) {
  return type.split('').map(function (char) {
    return ENCODING_MAP[char];
  }).join('');
}

function decode(encoded) {
  return encoded.split('').map(function (char) {
    return DECODING_MAP[char];
  }).join('');
}

function toGlobalID(type, pushID) {
  return pushID + '-' + encode(type);
}

function fromGlobalID(globalID) {
  (0, _invariant2.default)(isGlobalID(globalID), 'Invalid globalID.');
  return {
    type: decode(globalID.substring(21)),
    pushID: globalID.substring(0, 20)
  };
}

function newGlobalID(type) {
  return toGlobalID(type, generatePushID());
}

function typeFromGlobalID(globalID) {
  return fromGlobalID(globalID).type;
}

function isGlobalID(globalID) {
  return globalID && typeof globalID === 'string' ? globalID.match(/^[A-Za-z0-9-_]{20}-[A-Za-z0-9_]+$/) !== null : false;
}